"""
Pulling, Parsing, and Saving KEGG Entries
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Functionality for pulling KEGG entries from the KEGG REST API, parsing them, and saving the entries as files.
"""
import multiprocessing as mp
import os
import abc
import typing as t
import pickle as p
import zipfile as zf

from . import kegg_url as ku
from . import rest as r


class PullResult:
    """The collections of entry IDs, each of which resulted in a particular KEGG Response status after a pull."""

    def __init__(self) -> None:
        self._successful_entry_ids = []
        self._failed_entry_ids = []
        self._timed_out_entry_ids = []

    def __repr__(self):
        def get_entry_ids_repr(entry_ids) -> str:
            return ', '.join(entry_ids) if len(entry_ids) > 0 else 'none'

        successful_entry_ids: str = get_entry_ids_repr(entry_ids=self.successful_entry_ids)
        failed_entry_ids: str = get_entry_ids_repr(entry_ids=self.failed_entry_ids)
        timed_out_entry_ids: str = get_entry_ids_repr(entry_ids=self.timed_out_entry_ids)

        return f'Successful Entry Ids: {successful_entry_ids}\n' \
               f'Failed Entry Ids: {failed_entry_ids}\n' \
               f'Timed Out Entry Ids: {timed_out_entry_ids}'

    @property
    def successful_entry_ids(self) -> tuple:
        """The IDs of entries successfully pulled."""
        return tuple(self._successful_entry_ids)

    @property
    def failed_entry_ids(self) -> tuple:
        """The IDs of entries that failed to be pulled."""
        return tuple(self._failed_entry_ids)

    @property
    def timed_out_entry_ids(self) -> tuple:
        """The IDs of entries that timed out before being pulled."""
        return tuple(self._timed_out_entry_ids)

    def add_entry_ids(self, *entry_ids, status: r.KEGGresponse.Status) -> None:
        """ Adds entry IDs of a given status to the pull result.

        :param entry_ids: The entry IDs to add.
        :param status: The status resulting from attempting to pull the entries of the given IDs.
        """
        if status == r.KEGGresponse.Status.SUCCESS:
            self._successful_entry_ids.extend(entry_ids)
        elif status == r.KEGGresponse.Status.FAILED:
            self._failed_entry_ids.extend(entry_ids)
        else:
            self._timed_out_entry_ids.extend(entry_ids)

    def merge_pull_results(self, other) -> None:
        """ Combines two pull results into one.

        :param other: The pull result to combine into this one.
        """
        self._successful_entry_ids.extend(other.successful_entry_ids)
        self._failed_entry_ids.extend(other.failed_entry_ids)
        self._timed_out_entry_ids.extend(other.timed_out_entry_ids)


class SinglePull:
    """Class capable of performing a single request to the KEGG REST API for pulling up to a maximum number of entries."""

    class _AbstractEntrySaver(abc.ABC):
        """Abstract class for saving KEGG entries as files."""

        def save(self, entry_id: str, entry: t.Union[str, bytes], entry_field: str) -> None:
            """ Saves a KEGG entry as a file.

            :param entry_id: The entry ID (part of the file name).
            :param entry: The entry to save (contents of the file).
            :param entry_field: The particular field of the entry (file extension).
            """
            file_extension = 'txt' if entry_field is None else entry_field
            file_name = f'{entry_id}.{file_extension}'
            self._save(file_name=file_name, entry=entry)

        @abc.abstractmethod
        def _save(self, file_name: str, entry: t.Union[str, bytes]) -> None:
            """ Saves a KEGG entry in the particular manner of the child class.

            :param file_name: The name of the file in which the KEGG entry is saved.
            :param entry: The entry to save (contents of the file).
            """
            pass  # pragma: no cover

    class _DirectoryEntrySaver(_AbstractEntrySaver):
        """Class that saves KEGG entries in a directory."""

        def __init__(self, output_dir: str) -> None:
            """
            :param output_dir: The directory in which to save the KEGG entries.
            """
            if not os.path.isdir(output_dir):
                os.mkdir(output_dir)

            self._output_dir = output_dir

        def _save(self, file_name: str, entry: t.Union[str, bytes]) -> None:
            """ Saves a KEGG entry in a file within a directory.

            :param file_name: The name of the file to save in the directory.
            :param entry: The entry to save (content of the file).
            """
            file_path = os.path.join(self._output_dir, file_name)
            save_type = 'wb' if type(entry) is bytes else 'w'

            with open(file_path, save_type) as file:
                file.write(entry)

    class _ZipEntrySaver(_AbstractEntrySaver):
        """Class that saves KEGG entries in a ZIP file."""

        def __init__(self, zip_file: str) -> None:
            """
            :param zip_file: The path to the zip file to save the entries in.
            """
            self._zip_file = zip_file

        def _save(self, file_name: str, entry: t.Union[str, bytes]) -> None:
            """ Saves a KEGG entry in a zip file.

            :param file_name: The name of the file to save.
            :param entry: The entry to save (contents of file).
            """
            with zf.ZipFile(self._zip_file, 'a') as zip_file:
                zip_file.writestr(file_name, entry)

    def __init__(self, output_dir: str, kegg_rest: r.KEGGrest = None, entry_field: str = None) -> None:
        """
        :param output_dir: Path to the location where entries are saved (treated like a zip file if ends in .zip, else a directory).
        :param kegg_rest: Optional KEGGrest object used to make the requests to the KEGG REST API (a KEGGrest object with the default settings is created if one is not provided).
        :param entry_field: Optional KEGG entry field to pull.
        """
        if output_dir.endswith('.zip'):
            self._entry_saver = SinglePull._ZipEntrySaver(zip_file=output_dir)
        else:
            self._entry_saver = SinglePull._DirectoryEntrySaver(output_dir=output_dir)

        self._kegg_rest = kegg_rest if kegg_rest is not None else r.KEGGrest()
        self.entry_field = entry_field


    def pull(self, entry_ids: list) -> PullResult:
        """ Makes a single request to the KEGG REST API to pull one or more entries and save them as files.

        :param entry_ids: The IDs of the entries to pull and save.
        :return: The pull result.
        """
        kegg_response: r.KEGGresponse = self._kegg_rest.get(entry_ids=entry_ids, entry_field=self.entry_field)
        get_url: ku.GetKEGGurl = kegg_response.kegg_url
        pull_result = PullResult()

        if kegg_response.status == r.KEGGresponse.Status.SUCCESS:
            if get_url.multiple_entry_ids:
                self._save_multi_entry_response(kegg_response=kegg_response, pull_result=pull_result)
            else:
                self._save_single_entry_response(kegg_response=kegg_response, pull_result=pull_result)
        else:
            self._handle_unsuccessful_url(kegg_response=kegg_response, pull_result=pull_result)

        return pull_result

    def _save_multi_entry_response(self, kegg_response: r.KEGGresponse, pull_result: PullResult) -> None:
        """ Saves multiple entries within a KEGG response.

        :param kegg_response: The response from KEGG with multiple entries.
        :param pull_result: The pull result to update based on the status of the pull(s).
        """
        entries: list = self._separate_entries(concatenated_entries=kegg_response.text_body)
        get_url: ku.GetKEGGurl = kegg_response.kegg_url

        if len(entries) < len(get_url.entry_ids):
            # If we did not get all the entries requested, process each entry one at a time
            self._pull_separate_entries(get_url=get_url, pull_result=pull_result)
        else:
            pull_result.add_entry_ids(*get_url.entry_ids, status=r.KEGGresponse.Status.SUCCESS)

            for entry_id, entry in zip(get_url.entry_ids, entries):
                self._entry_saver.save(entry_id=entry_id, entry=entry, entry_field=self.entry_field)

    def _separate_entries(self, concatenated_entries: str) -> list:
        """ Separates the entries in a multi-entry response from KEGG.

        :param concatenated_entries: The response body with all the entries together.
        :return: The separated KEGG entries.
        """
        field_to_separator = {
            'aaseq': SinglePull._gene_separator, 'kcf': SinglePull._standard_separator,
            'mol': SinglePull._mol_separator, 'ntseq': SinglePull._gene_separator
        }

        if self.entry_field is None:
            separator = SinglePull._standard_separator
        else:
            separator = field_to_separator[self.entry_field]

        entries: list = separator(concatenated_entries=concatenated_entries)
        entries = [entry.strip() for entry in entries]

        return entries

    @staticmethod
    def _gene_separator(concatenated_entries: str) -> list:
        """ Separates KEGG entries that are separated by a deliminator for nucleotide and amino acid sequence entries.
        :param concatenated_entries: The response body with the gene-related entries together.
        :return: The separated gene-related entries.
        """
        entries: list = concatenated_entries.split('>')

        if len(entries) > 1:
            return entries[1:]
        else:
            return entries

    @staticmethod
    def _mol_separator(concatenated_entries: str) -> list:
        """ Separates mol entries.

        :param concatenated_entries: The response body with the mol entries together.
        :return: The separated mole entries.
        """
        return SinglePull._split_and_remove_last(concatenated_entries=concatenated_entries, deliminator='$$$$')

    @staticmethod
    def _split_and_remove_last(concatenated_entries: str, deliminator: str) -> list:
        """ Splits entries based on a deliminator and removes the resulting empty string at the end of the list.

        :param concatenated_entries: The response body with the entries together.
        :param deliminator: The deliminator that separates the entries and is at the end of the response body.
        :return: The separated entries with the empty string removed.
        """
        entries: list = concatenated_entries.split(deliminator)

        if len(entries) > 1:
            return entries[:-1]
        else:
            return entries

    @staticmethod
    def _standard_separator(concatenated_entries: str) -> list:
        """ The separation method for most types of KEGG entries.

        :param concatenated_entries: The response body with the KEGG entries together.
        :return: The separated entries.
        """
        return SinglePull._split_and_remove_last(concatenated_entries=concatenated_entries, deliminator='///')

    def _pull_separate_entries(self, get_url: ku.GetKEGGurl, pull_result: PullResult) -> None:
        """ Pulls one entry at a time for a Get KEGG URL that has multiple entry IDs.

        :param get_url: The Get KEGG URL with multiple entry IDs to pull one at a time.
        :param pull_result: The pull result to update based on the success of the pulling.
        """
        for entry_id in get_url.entry_ids:
            kegg_response: r.KEGGresponse = self._kegg_rest.get(entry_ids=[entry_id], entry_field=self.entry_field)

            if kegg_response.status == r.KEGGresponse.Status.SUCCESS:
                self._save_single_entry_response(kegg_response=kegg_response, pull_result=pull_result)
            else:
                pull_result.add_entry_ids(entry_id, status=kegg_response.status)

    def _save_single_entry_response(self, kegg_response: r.KEGGresponse, pull_result: PullResult) -> None:
        """ Saves the entry in a KEGG response that contains only one entry.

        :param kegg_response: The KEGG response that only has one entry.
        :param pull_result: The pull result to update with the successful entry ID.
        """
        get_url: ku.GetKEGGurl = kegg_response.kegg_url
        [entry_id] = get_url.entry_ids
        pull_result.add_entry_ids(entry_id, status=r.KEGGresponse.Status.SUCCESS)

        if ku.GetKEGGurl.is_binary(entry_field=self.entry_field):
            entry: bytes = kegg_response.binary_body
        else:
            entry: str = kegg_response.text_body

        self._entry_saver.save(entry_id=entry_id, entry=entry, entry_field=self.entry_field)


    def _handle_unsuccessful_url(self, kegg_response: r.KEGGresponse, pull_result: PullResult) -> None:
        """ Handles an unsuccessful pull (failed or timed out).

        :param kegg_response: The KEGG response resulting from a pull that was not successful.
        :param pull_result: The pull result to update based on how the unsuccessful response is dealt with.
        """
        get_url: ku.GetKEGGurl = kegg_response.kegg_url
        status: r.KEGGresponse.Status = kegg_response.status

        if get_url.multiple_entry_ids:
            self._pull_separate_entries(get_url=get_url, pull_result=pull_result)
        else:
            [entry_id] = get_url.entry_ids
            pull_result.add_entry_ids(entry_id, status=status)


class AbstractMultiplePull(abc.ABC):
    """Abstract class that makes multiple requests to the KEGG REST API to pull and save entries of an unlimited amount."""

    def __init__(self, single_pull: SinglePull, force_single_entry: bool = False) -> None:
        """
        :param single_pull: The SinglePull object used for each pull.
        :param force_single_entry: Determines whether to pull only one entry at a time regardless of the entry field specified in the SinglePull argument.
        """
        self._single_pull = single_pull
        self._force_single_entry = force_single_entry

    def pull(
        self, entry_ids: list, entry_field: t.Optional[str] = '', force_single_entry: t.Optional[bool] = None
    ) -> PullResult:
        """ Makes multiple requests to the KEGG REST API for an unlimited amount of entry IDs.

        :param entry_ids: The IDs of the entries that are split into multiple pulls.
        :param entry_field: An optional field of the entries to pull.
        :param force_single_entry: If provided, updates the force_single_single_entry value of this AbstractMultiplePull object for this call of pull.
        :return: The pull result.
        """
        force_single_entry = force_single_entry if force_single_entry is not None else self._force_single_entry

        # If a new entry field is provided, update it
        if entry_field != '':
            self._single_pull.entry_field = entry_field

        entry_ids: list = self._group_entry_ids(
            entry_ids_to_group=entry_ids, force_single_entry=force_single_entry
        )

        return self._pull(grouped_entry_ids=entry_ids)

    def _group_entry_ids(self, entry_ids_to_group: list, force_single_entry: bool) -> list:
        group_size: int = AbstractMultiplePull._get_n_entries_per_url(
            force_single_entry=force_single_entry, entry_field=self._single_pull.entry_field
        )

        grouped_entry_ids = []

        for i in range(0, len(entry_ids_to_group), group_size):
            entry_id_group: list = entry_ids_to_group[i:i + group_size]
            grouped_entry_ids.append(entry_id_group)

        return grouped_entry_ids

    @staticmethod
    def _get_n_entries_per_url(force_single_entry: bool, entry_field: str) -> int:
        """ Gets the number of entries for each Get KEGG URL used to make the pulls.

        :param force_single_entry: If true, returns 1.
        :param entry_field: If the entry field can only be pulled 1 at a time, return 1.
        :return: 1 if the above specifications are true, else the maximum number of entries per Get URL.
        """
        if force_single_entry:
            return 1
        elif ku.GetKEGGurl.only_one_entry(entry_field=entry_field):
            return 1
        else:
            return ku.GetKEGGurl.MAX_ENTRY_IDS_PER_URL

    @abc.abstractmethod
    def _pull(self, grouped_entry_ids: list) -> PullResult:
        """ Pulls the entries of specified IDs in the manner of the extended concrete class.

        :param grouped_entry_ids: List of lists of entry IDs, with each list being below or equal to the number allowed per Get KEGG URL.
        :return: The pull result.
        """
        pass  # pragma: no cover


class SingleProcessMultiplePull(AbstractMultiplePull):
    """Class that makes multiple requests to the KEGG REST API to pull entries within a single process."""

    def _pull(self, grouped_entry_ids: list) -> PullResult:
        """ Makes multiple requests to the KEGG REST API to pull entries within a single process.

        :param grouped_entry_ids: List of lists of entry IDs, with each list being below or equal to the number allowed per Get KEGG URL.
        :return: The pull result
        """
        multiple_pull_result = PullResult()

        for entry_id_group in grouped_entry_ids:
            single_pull_result: PullResult = self._single_pull.pull(entry_ids=entry_id_group)
            multiple_pull_result.merge_pull_results(other=single_pull_result)

        return multiple_pull_result


class MultiProcessMultiplePull(AbstractMultiplePull):
    """Class that makes multiple requests to the KEGG REST API to pull entries within multiple processes."""

    def __init__(self, single_pull: SinglePull, n_workers: int, force_single_entry: bool = False):
        """
        :param single_pull: The SinglePull object used for each pull.
        :param force_single_entry: Determines whether to pull only one entry at a time despite the entry field specified in the SinglePull argument.
        :param n_workers: The number of processes to use.
        """
        super(MultiProcessMultiplePull, self).__init__(single_pull=single_pull, force_single_entry=force_single_entry)
        self._n_workers = n_workers if n_workers is not None else os.cpu_count()

    def _pull(self, grouped_entry_ids: list) -> PullResult:
        """ Makes multiple requests to the KEGG REST API to pull entries within multiple processes.

        :param grouped_entry_ids: List of lists of entry IDs, with each list being below or equal to the number allowed per Get KEGG URL.
        :return: The pull result
        """
        multiple_pull_result = PullResult()
        args = [(entry_ids, self._single_pull) for entry_ids in grouped_entry_ids]
        chunk_size: int = min(len(grouped_entry_ids) // self._n_workers, 10)

        if chunk_size == 0:
            chunk_size = 1

        with mp.Pool(self._n_workers) as pool:
            results: list = pool.starmap(_get_single_pull_result, args, chunksize=chunk_size)

        for result in results:
            single_pull_result: PullResult = p.loads(result)
            multiple_pull_result.merge_pull_results(other=single_pull_result)

        return multiple_pull_result


def _get_single_pull_result(entry_ids: list, single_pull: SinglePull) -> bytes:
    """ Makes a request to the REST KEGG API to pull one or more entries.

    :param entry_ids: The IDs of the entries to pull.
    :param single_pull: The SinglePull object used to make the pull.
    :return: The pull result.
    """
    single_pull_result: PullResult = single_pull.pull(entry_ids=entry_ids)

    return p.dumps(single_pull_result)



"""
Constructing URLs for the KEGG REST API
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Classes for creating and validating KEGG REST API URLs.
"""
import requests as rq
import logging as l
import abc
import typing as t

from . import _utils as u

BASE_URL: str = 'https://rest.kegg.jp'


class AbstractKEGGurl(abc.ABC):
    """Abstract class which validates and constructs URLs for accessing the KEGG REST API and contains the base data and functionality for all KEGG URL classes."""

    _valid_kegg_databases = {
        'pathway', 'brite', 'module', 'ko', 'genome', 'vg', 'vp', 'ag', 'compound', 'glycan', 'reaction', 'rclass',
        'enzyme', 'network', 'variant', 'disease', 'drug', 'dgroup', 'genes', 'ligand', 'kegg'
    }

    _valid_medicus_databases = {
        'disease_ja', 'drug_ja', 'dgroup_ja', 'compound_ja', 'brite_ja', 'atc', 'jtc', 'ndc', 'yj'
    }

    _organism_set = None

    def __init__(self, rest_operation: str, base_url: str = BASE_URL, **kwargs) -> None:
        """
        :param rest_operation: The KEGG REST API operation in the URL.
        :param base_url: The base URL for accessing the KEGG web API.
        :param kwargs: The arguments used to construct the REST options after they are validated.
        :raises ValueError: Raised if the given arguments cannot construct a valid KEGG URL.
        """
        self._validate(**kwargs)
        url_options: str = self._create_rest_options(**kwargs)
        self._url = f'{base_url}/{rest_operation}/{url_options}'

    # noinspection PyMethodParameters
    @u.staticproperty
    def organism_set() -> set:
        """ Obtains the set of valid KEGG organism database names by requesting from the KEGG REST API (caches this result so the request only needs to be done once).

        :return: The set of organism database names.
        :raises RuntimeError: Raised in the unlikely case that the request fails.
        """
        if AbstractKEGGurl._organism_set is None:
            url = f'{BASE_URL}/list/organism'
            error_message = 'The request to the KEGG web API {} while fetching the organism set using the URL: {}'

            try:
                response: rq.Response = rq.get(url=url, timeout=60)
            except rq.exceptions.Timeout:
                raise RuntimeError(
                    error_message.format('timed out', url)
                )

            status_code: int = response.status_code

            if status_code != 200:
                raise RuntimeError(
                    error_message.format(f'failed with status code {status_code}', url)
                )

            organism_list: list = response.text.strip().split('\n')
            AbstractKEGGurl._organism_set = set()

            for organism in organism_list:
                [code, name, _, _] = organism.strip().split('\t')
                AbstractKEGGurl._organism_set.add(code)
                AbstractKEGGurl._organism_set.add(name)

        return AbstractKEGGurl._organism_set

    @abc.abstractmethod
    def _validate(self, **kwargs) -> None:
        """ Ensures the arguments passed into the constructor result in a valid KEGG URL.

        :param kwargs: The arguments to validate.
        :raises ValueError: Raised if the given arguments cannot construct a valid KEGG URL.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def _create_rest_options(self, **kwargs) -> str:
        """ Creates the string at the end part of a KEGG URL to specify the options for a REST API request.

        :param kwargs: The arguments used to create the options.
        :return: The REST API options.
        """
        pass  # pragma: no cover

    @property
    def url(self) -> str:
        """The constructed and validated KEGG URL."""
        return self._url

    def __repr__(self) -> str:
        return self.url

    @staticmethod
    def _raise_error(reason: str) -> None:
        """ Raises an exception for when a URL is not valid.

        :param reason: The reason why the URL was not valid.
        :raises ValueError: The error that is raised.
        """
        raise ValueError(f'Cannot create URL - {reason}')

    @staticmethod
    def _validate_rest_option(
        option_name: str, option_value: str, valid_rest_options: t.Iterable, add_org: bool = False
    ) -> None:
        """ Raises an exception if a provided REST API option is not valid.

        :param option_name: The name of the type of option to check.
        :param option_value: The value of the REST API option provided.
        :param valid_rest_options: The collection of valid options to choose from.
        :param add_org: Whether to add the "<org>" option to the valid options in the error message.
        :raises ValueError: Raised when the provided option is not valid.
        """
        if option_value not in valid_rest_options:
            if add_org:
                valid_rest_options: set = set(valid_rest_options)
                valid_rest_options.add('<org>')

            valid_options = ', '.join(sorted(valid_rest_options))
            error_reason = f'Invalid {option_name}: "{option_value}". Valid values are: {valid_options}.'

            if add_org:
                error_reason += ' Where <org> is an organism code or T number.'

            AbstractKEGGurl._raise_error(reason=error_reason)

    @staticmethod
    def _validate_database_name(database_name: str, extra_databases: set = set(), excluded_databases: set = set()):
        """ Ensures the database provided is a valid KEGG database.

        :param database_name: The name of the database to validate.
        :param extra_databases: Additional optional database names to add to the core KEGG databases for the validation.
        :param excluded_databases: Optional database names to exclude from the validation. If extra_databases overlaps
        excluded_databases, extra_databases has priority.
        :raises ValueError: Raised when the provided database is not valid.
        """
        if database_name not in AbstractKEGGurl.organism_set:
            valid_databases = AbstractKEGGurl._valid_kegg_databases.union(AbstractKEGGurl._valid_medicus_databases)
            valid_databases: set = valid_databases - excluded_databases
            valid_databases: set = valid_databases.union(extra_databases)

            AbstractKEGGurl._validate_rest_option(
                option_name='database name', option_value=database_name, valid_rest_options=valid_databases,
                add_org=True
            )


class ListKEGGurl(AbstractKEGGurl):
    """Contains URL construction and validation functionality of the KEGG API list operation."""

    def __init__(self, database_name: str):
        """
        :param database_name: The database option for the KEGG list URL.
        :raises ValueError: Raised if the provided database is not valid.
        """
        super().__init__(rest_operation='list', database_name=database_name)

    def _validate(self, database_name: str):
        """ Ensures the database option is a KEGG database supported by the list operation.

        :param database_name: The name of the database to check.
        :raises ValueError: Raised if the provided database is not valid.
        """
        AbstractKEGGurl._validate_database_name(
            database_name=database_name, extra_databases={'organism'}, excluded_databases={'genes', 'ligand', 'kegg'}
        )

    def _create_rest_options(self, database_name: str) -> str:
        """ Implements the KEGG REST API options creation by returning the provided database name (the only option).

        :param database_name: The database option to return.
        :return: The database option.
        """
        return database_name


class InfoKEGGurl(AbstractKEGGurl):
    """Contains URL construction and validation functionality of the KEGG API info operation."""

    def __init__(self, database_name: str) -> None:
        """
        :param database_name: The database option for the KEGG info URL.
        :raises ValueError: Raised if the provided database is not valid.
        """
        super(InfoKEGGurl, self).__init__(rest_operation='info', database_name=database_name)

    def _validate(self, database_name: str):
        """ Ensures the database option is a KEGG database supported by the info operation.

        :param database_name: The name of the database to check.
        :raises ValueError: Raised if the provided database is not valid.
        """
        AbstractKEGGurl._validate_database_name(
            database_name=database_name, excluded_databases=AbstractKEGGurl._valid_medicus_databases
        )

    def _create_rest_options(self, database_name: str) -> str:
        """ Implements the KEGG REST API options creation by returning the provided database name (the only option).

        :param database_name: The database option to return.
        :return: The database option.
        """
        return database_name


class GetKEGGurl(AbstractKEGGurl):
    """Contains URL construction and validation functionality for the KEGG API get operation."""

    _entry_fields = {
        'aaseq': True, 'ntseq': True, 'mol': True, 'kcf': True, 'image': False, 'conf': False, 'kgml': False,
        'json': False
    }

    _MAX_ENTRY_IDS_PER_URL = 10

    # noinspection PyMethodParameters
    @u.staticproperty
    def MAX_ENTRY_IDS_PER_URL() -> int:
        """ The maximum number of entry IDs allowed in a single get KEGG URL."""
        return GetKEGGurl._MAX_ENTRY_IDS_PER_URL

    def __init__(self, entry_ids: list, entry_field: str = None) -> None:
        """
        :param entry_ids: Specifies which entry IDs go in the first option of the URL.
        :param entry_field: Specifies which entry field goes in the second option.
        :raises ValueError: Raised if the entry IDs or entry field is not valid.
        """
        super().__init__(rest_operation='get', entry_ids=entry_ids, entry_field=entry_field)
        self._entry_ids = entry_ids
        self._entry_field = entry_field

    @property
    def entry_ids(self) -> list:
        """The entry IDs of the get KEGG URL."""
        return self._entry_ids

    @property
    def multiple_entry_ids(self) -> bool:
        """Determines whether the get KEGG URL has more than one entry ID."""
        return len(self.entry_ids) > 1

    def _validate(self, entry_ids: list, entry_field: str):
        """ Ensures valid Entry IDs and a valid entry field are provided.

        :param entry_ids: The entry IDs to validate.
        :param entry_field: The entry field to validate.
        :raises ValueError: Raised if the entry IDs or entry field is not valid.
        """
        n_entry_ids: int = len(entry_ids)

        if n_entry_ids == 0:
            self._raise_error(reason='Entry IDs must be specified for the KEGG get operation')

        max_entry_ids: int = GetKEGGurl._MAX_ENTRY_IDS_PER_URL

        if n_entry_ids > max_entry_ids:
            self._raise_error(
                reason=f'The maximum number of entry IDs is {max_entry_ids} but {n_entry_ids} were provided'
            )

        if entry_field is not None:
            AbstractKEGGurl._validate_rest_option(
                option_name='KEGG entry field', option_value=entry_field, valid_rest_options=GetKEGGurl._entry_fields
            )

            if self.only_one_entry(entry_field=entry_field) and n_entry_ids > 1:
                self._raise_error(
                    reason=f'The KEGG entry field: "{entry_field}" only supports requests of one KEGG entry '
                           f'at a time but {n_entry_ids} entry IDs are provided'
                )

    @staticmethod
    def only_one_entry(entry_field: str) -> bool:
        """ Determines whether a KEGG entry field can only be pulled in one entry at a time for the KEGG get API
        operation.

        :param entry_field: The KEGG entry field to check.
        """
        return entry_field is not None and not GetKEGGurl._entry_fields[entry_field]

    @staticmethod
    def is_binary(entry_field: str) -> bool:
        """ Determines if the entry field is a binary response or not.

        :param entry_field: The KEGG entry field to check.
        """
        return entry_field == 'image'

    def _create_rest_options(self, entry_ids: list, entry_field: str) -> str:
        """ Constructs the REST options for the KEGG API get operation.

        :param entry_ids: The entry IDs for the first REST option.
        :param entry_field: The entry field for the second REST option.
        :return: The constructed options.
        """
        entry_ids_url_option = '+'.join(entry_ids)

        if entry_field is not None:
            return f'{entry_ids_url_option}/{entry_field}'
        else:
            return entry_ids_url_option


class KeywordsFindKEGGurl(AbstractKEGGurl):
    """Contains the URL construction and validation functionality for the KEGG API find operation based on the URL form that searches entries by keywords."""

    def __init__(self, database_name: str, keywords: list):
        """
        :param database_name: The database name option for the first part of the URL.
        :param keywords: The keyword options for the second part of the URL.
        :raises ValueError: Raised if the database name is invalid or keywords are not provided.
        """
        super(KeywordsFindKEGGurl, self).__init__(rest_operation='find', database_name=database_name, keywords=keywords)

    def _validate(self, database_name: str, keywords: list):
        """ Ensures keywords are provided and the database name is valid.

        :param database_name: The database name to check.
        :param keywords: The keywords to check.
        :raises ValueError: Raised if the database name is invalid or keywords are not provided.
        """
        if len(keywords) == 0:
            self._raise_error(reason='No search keywords specified')

        AbstractKEGGurl._validate_database_name(database_name=database_name, excluded_databases={'brite', 'kegg'})

    def _create_rest_options(self, keywords: list, database_name: str) -> str:
        """ Constructs the options for the URL using the database name followed by the keywords.

        :param keywords: The keywords to go in the options.
        :param database_name: The database name to go in the options.
        :return: The constructed options.
        """
        keywords_string = '+'.join(keywords)

        return f'{database_name}/{keywords_string}'


class MolecularFindKEGGurl(AbstractKEGGurl):
    """Contains the URL construction and validation functionality for the KEGG API find operation based on the URL form that uses chemical / molecular attributes of compounds."""

    _valid_molecular_databases = {'compound', 'drug'}

    def __init__(self, database_name: str, formula: str = None, exact_mass: t.Union[float, tuple] = None,
        molecular_weight: t.Union[int, tuple] = None
    ):
        """
        :param database_name: The database name option for the first part of the URL.
        :param formula: The chemical formula option that can go in the second part of the URL.
        :param exact_mass: The exact molecule mass option that can go in the second part of the URL.
        :param molecular_weight: The molecular weight option that can go in the second part of the URL.
        :raises ValueError: Raised if the provided database name or molecular attribute is invalid.
        """
        super(MolecularFindKEGGurl, self).__init__(
            rest_operation='find', database_name=database_name, formula=formula, exact_mass=exact_mass,
            molecular_weight=molecular_weight
        )

    def _validate(
        self, database_name: str, formula: str = None, exact_mass: t.Union[float, tuple] = None,
        molecular_weight: t.Union[int, tuple] = None
    ):
        """ Ensures a valid database name and molecular attributes are provided.

        :param database_name: The database name to check.
        :param formula: The chemical formula attribute to check.
        :param exact_mass: The exact mass attribute to check.
        :param molecular_weight: The molecular weight attribute to check.
        :raises ValueError: Raised if the provided database name or molecular attribute is invalid.
        """
        AbstractKEGGurl._validate_rest_option(
            option_name='molecular database name', option_value=database_name,
            valid_rest_options=MolecularFindKEGGurl._valid_molecular_databases
        )

        if formula is None and exact_mass is None and molecular_weight is None:
            AbstractKEGGurl._raise_error(
                reason='Must provide either a chemical formula, exact mass, or molecular weight option'
            )

        if formula is not None and (exact_mass is not None or molecular_weight is not None):
            l.warning(
                'Only a chemical formula, exact mass, or molecular weight is used to construct the URL. Using formula'
                '...'
            )
        elif formula is None and exact_mass is not None and molecular_weight is not None:
            l.warning('Both an exact mass and molecular weight are provided. Using exact mass...')

        MolecularFindKEGGurl._validate_range(range_values=exact_mass, range_name='Exact mass')
        MolecularFindKEGGurl._validate_range(range_values=molecular_weight, range_name='Molecular weight')

    @staticmethod
    def _validate_range(range_values: tuple, range_name: str):
        """ Ensures a given range is valid.

        :param range_values: The two end points of the range (start and end).
        :param range_name: The name of the range for the resulting error message in case of an invalid range.
        :raises ValueError: Raised if the range values are not valid.
        """
        if range_values is not None and type(range_values) is tuple:
            if len(range_values) != 2:
                provided_values = ', '.join(str(range_value) for range_value in range_values)

                AbstractKEGGurl._raise_error(
                    f'{range_name} range can only be constructed from 2 values but {len(range_values)} are provided: '
                    f'{provided_values}'
                )

            min_val, max_val = range_values

            if not min_val < max_val:
                AbstractKEGGurl._raise_error(
                    reason=f'The first value in the range must be less than the second. Values provided:'
                           f' {min_val}-{max_val}'
                )

    def _create_rest_options(self, database_name: str, formula: str = None, exact_mass: t.Union[float, tuple] = None,
        molecular_weight: t.Union[int, tuple] = None
    ) -> str:
        """ Constructs the options for the URL using the database name followed by a molecular attribute.

        :param database_name: The database name option in the first part of the URL.
        :param formula: The chemical formula option that can go in the second part of the URL.
        :param exact_mass: The exact mass attribute that can go in the second part of the URL.
        :param molecular_weight: The molecular weight attribute that can go in the second part of the URL.
        :return: The constructed options.
        """
        if formula is not None:
            options = f'{formula}/formula'
        elif exact_mass is not None:
            options = MolecularFindKEGGurl._get_range_options(option_name='exact_mass', option_value=exact_mass)
        else:
            options = MolecularFindKEGGurl._get_range_options(option_name='mol_weight', option_value=molecular_weight)

        return f'{database_name}/{options}'

    @staticmethod
    def _get_range_options(option_name: str, option_value: t.Union[float, int, tuple]) -> str:
        """ Constructs options for the URL that are either a single number or a range (start and end separated by a
        dash).

        :param option_name: The name of the option to go in the third part of the URL.
        :param option_value: The single number or range.
        :return: The constructed option.
        """
        if type(option_value) is int or type(option_value) is float:
            options = option_value
        else:
            minimum, maximum = option_value

            options = f'{minimum}-{maximum}'

        return f'{options}/{option_name}'


class AbstractConvKEGGurl(AbstractKEGGurl):
    """Abstract class containing data shared by the KEGG URL classes that validate and construct URLs for the conv KEGG
    REST API operation."""

    _valid_outside_gene_databases = {'ncbi-geneid', 'ncbi-proteinid', 'uniprot'}
    _valid_kegg_molecule_databases = {'compound', 'glycan', 'drug'}
    _valid_outside_molecule_databases = {'pubchem', 'chebi'}

    def __init__(self, **kwargs):
        """
        :param kwargs: Arguments for the URL validation and construction.
        :raises ValueError: Raised if the provided arguments cannot construct a valid conv KEGG URL.
        """
        super(AbstractConvKEGGurl, self).__init__(rest_operation='conv', **kwargs)

    @abc.abstractmethod
    def _validate(self, **kwargs):
        """ Validates options for a conv KEGG URL.

        :param kwargs: The options to validate.
        :raises ValueError: Raised if the provided arguments cannot construct a valid conv KEGG URL.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def _create_rest_options(self, **kwargs) -> str:
        """ Constructs the options in a conv KEGG URL.

        :param kwargs: The arguments to create the options.
        :return: The constructed options.
        """
        pass  # pragma: no cover


class DatabaseConvKEGGurl(AbstractConvKEGGurl):
    """Contains the URL construction and validation functionality of the KEGG API conv operation based on the URL form that uses a KEGG database and an outside database."""

    def __init__(self, kegg_database_name: str, outside_database_name: str):
        """
        :param kegg_database_name: The name of the KEGG database.
        :param outside_database_name: The name of the outside database.
        :raises ValueError: Raised if the database names are not valid or are not of the same type.
        """
        super(DatabaseConvKEGGurl, self).__init__(
            kegg_database_name=kegg_database_name, outside_database_name=outside_database_name
        )

    def _validate(self, kegg_database_name: str, outside_database_name: str):
        """ Ensures that the database names are valid and that they're both the same type

        :param kegg_database_name: The name of the KEGG database to check.
        :param outside_database_name: The name of the outside database to check.
        :raises ValueError: Raised if the database names are not valid or are not of the same type.
        """
        valid_kegg_gene_databases: set = AbstractKEGGurl.organism_set
        valid_kegg_molecule_databases: set = AbstractConvKEGGurl._valid_kegg_molecule_databases
        valid_kegg_databases: set = valid_kegg_molecule_databases.union(valid_kegg_gene_databases)

        if kegg_database_name not in valid_kegg_databases:
            AbstractKEGGurl._validate_rest_option(
                option_name='KEGG database', option_value=kegg_database_name,
                valid_rest_options=valid_kegg_molecule_databases, add_org=True
            )

        valid_outside_gene_databases: set = AbstractConvKEGGurl._valid_outside_gene_databases
        valid_outside_molecule_databases: set = AbstractConvKEGGurl._valid_outside_molecule_databases
        valid_outside_databases: set = valid_outside_molecule_databases.union(valid_outside_gene_databases)

        AbstractKEGGurl._validate_rest_option(
            option_name='outside database', option_value=outside_database_name,
            valid_rest_options=valid_outside_databases
        )

        if kegg_database_name in valid_kegg_gene_databases and \
                outside_database_name not in valid_outside_gene_databases:
            AbstractKEGGurl._raise_error(
                reason=f'KEGG database "{kegg_database_name}" is a gene database but outside database '
                       f'"{outside_database_name}" is not.'
            )

        if kegg_database_name in valid_kegg_molecule_databases and \
                outside_database_name not in valid_outside_molecule_databases:
            AbstractKEGGurl._raise_error(
                reason=f'KEGG database "{kegg_database_name}" is a molecule database but outside database '
                       f'"{outside_database_name}" is not.'
            )

    def _create_rest_options(self, kegg_database_name: str, outside_database_name: str) -> str:
        """ Constructs the REST options by appending the outside database name to the kegg database name

        :param kegg_database_name: The KEGG database option.
        :param outside_database_name: The outside database option
        :return: The constructed options.
        """
        return f'{kegg_database_name}/{outside_database_name}'


class EntriesConvKEGGurl(AbstractConvKEGGurl):
    """Contains the URL construction and validation functionality for the KEGG API conv operation based on the URL form that uses a target database and entry IDs."""

    def __init__(self, target_database_name: str, entry_ids: list):
        """
        :param target_database_name: The target database option.
        :param entry_ids: The entry IDs options.
        :raises ValueError: Raised if the target database is invalid or entry IDs are not provided.
        """
        super(EntriesConvKEGGurl, self).__init__(target_database_name=target_database_name, entry_ids=entry_ids)

    def _validate(self, target_database_name: str, entry_ids: list):
        """ Ensures the target database is valid and that the entry IDs are provided.

        :param target_database_name: The name of the target database to check.
        :param entry_ids: The entry IDs to check.
        :raises ValueError: Raised if the target database is invalid or entry IDs are not provided.
        """
        valid_databases: set = AbstractConvKEGGurl._valid_kegg_molecule_databases
        valid_databases: set = valid_databases.union(AbstractConvKEGGurl._valid_outside_gene_databases)
        valid_databases: set = valid_databases.union(AbstractConvKEGGurl._valid_outside_molecule_databases)
        valid_databases.add('genes')

        if target_database_name not in valid_databases.union(AbstractKEGGurl.organism_set):
            AbstractKEGGurl._validate_rest_option(
                option_name='target database', option_value=target_database_name, valid_rest_options=valid_databases,
                add_org=True
            )

        if len(entry_ids) == 0:
            self._raise_error(reason='Entry IDs must be specified for this KEGG "conv" operation')

    def _create_rest_options(self, target_database_name: str, entry_ids: list) -> str:
        """ Constructs the REST options by appending the entry IDs (separated by '+') to the target database name.

        :param target_database_name: The name of the target database in the first part of the URL.
        :param entry_ids: The entry IDs in the second part of the URL.
        :return: The constructed options.
        """
        return f'{target_database_name}/{"+".join(entry_ids)}'


class AbstractLinkKEGGurl(AbstractKEGGurl):
    """Abstract class containing the shared data for the link KEGG URLs."""

    _extra_database_names = {'atc', 'jtc', 'ndc', 'yj', 'pubmed'}

    def __init__(self, **kwargs):
        """
        :param kwargs: The arguments to validate and construct the URL.
        :raises ValueError: Raised if the provided arguments are invalid.
        """
        super(AbstractLinkKEGGurl, self).__init__(rest_operation='link', **kwargs)

    @abc.abstractmethod
    def _validate(self, **kwargs):
        pass  # pragma: no cover

    @abc.abstractmethod
    def _create_rest_options(self, **kwargs) -> str:
        pass  # pragma: no cover


class DatabaseLinkKEGGurl(AbstractLinkKEGGurl):
    """Contains the URL construction and validation functionality for the link KEGG REST API operation of the form that uses a target database and a source database."""

    def __init__(self, target_database_name: str, source_database_name: str):
        """
        :param target_database_name: The name of the target database option.
        :param source_database_name: The name of the source database option.
        :raises ValueError: Raised if the databases are invalid.
        """
        super(DatabaseLinkKEGGurl, self).__init__(
            target_database_name=target_database_name, source_database_name=source_database_name
        )

    def _validate(self, target_database_name: str, source_database_name: str):
        """ Ensures the provided databases are valid

        :param target_database_name: The name of the target database to check.
        :param source_database_name: The name of the source database to check.
        :raises ValueError: Raised if the databases are invalid.
        """
        excluded_databases: set = AbstractKEGGurl._valid_medicus_databases.union({'kegg', 'genes', 'ligand'})

        AbstractKEGGurl._validate_database_name(
            database_name=target_database_name, extra_databases=AbstractLinkKEGGurl._extra_database_names,
            excluded_databases=excluded_databases
        )

        AbstractKEGGurl._validate_database_name(
            database_name=source_database_name, extra_databases=AbstractLinkKEGGurl._extra_database_names,
            excluded_databases=excluded_databases
        )

    def _create_rest_options(self, target_database_name: str, source_database_name: str) -> str:
        """ Constructs the options by appending the target database name to the source database name
        :param target_database_name: The target database name for the first option.
        :param source_database_name: The source database name for the second option.
        :return: The constructed options.
        """
        return f'{target_database_name}/{source_database_name}'


class EntriesLinkKEGGurl(AbstractLinkKEGGurl):
    """Contains the URL construction and validation functionality for the link KEGG REST API operation of the form that uses a target database and entry IDs."""

    def __init__(self, target_database_name: str, entry_ids: list):
        """
        :param target_database_name: The name of the target database option.
        :param entry_ids: The entry IDs options.
        :raises ValueError: Raised if the target database is invalid or entry IDs are not provided.
        """
        super(EntriesLinkKEGGurl, self).__init__(target_database_name=target_database_name, entry_ids=entry_ids)

    def _validate(self, target_database_name: str, entry_ids: list):
        """ Ensures the target database name is valid and that the entry IDs are provided.

        :param target_database_name: The name of the target database to check.
        :param entry_ids: The entry IDs to check.
        :raises ValueError: Raised if the target database is invalid or entry IDs are not provided.
        """
        excluded_databases: set = AbstractKEGGurl._valid_medicus_databases.union({'kegg', 'ligand'})

        AbstractKEGGurl._validate_database_name(
            database_name=target_database_name, extra_databases=AbstractLinkKEGGurl._extra_database_names,
            excluded_databases=excluded_databases
        )

        if len(entry_ids) == 0:
            AbstractKEGGurl._raise_error(reason='At least one entry ID must be specified to perform the link operation')

    def _create_rest_options(self, target_database_name: str, entry_ids: list) -> str:
        """Constructs the options by appending the entry IDs (separated by '+') to the target database name.

        :param target_database_name: The name of the target database for the first options.
        :param entry_ids: The entry IDs as the last options.
        :return: The constructed options.
        """
        return f'{target_database_name}/{"+".join(entry_ids)}'


class DdiKEGGurl(AbstractKEGGurl):
    """Contains the URL construction and validation functionality for the ddi KEGG REST operation."""

    def __init__(self, drug_entry_ids: list):
        """
        :param drug_entry_ids: The entry IDs for a drug database.
        :raises ValueError: Raised if the drug entry IDs are not provided.
        """
        super(DdiKEGGurl, self).__init__(rest_operation='ddi', drug_entry_ids=drug_entry_ids)

    def _validate(self, drug_entry_ids: list):
        """ Ensures the drug entry IDs are provided.

        :param drug_entry_ids: The drug entry IDs to check.
        :raises ValueError: Raised if the drug entry IDs are not provided.
        """
        if len(drug_entry_ids) == 0:
            AbstractKEGGurl._raise_error(reason='At least one drug entry ID must be specified for the DDI operation')

    def _create_rest_options(self, drug_entry_ids: list) -> str:
        """ Constructs the options by separating the drug entry IDs by '+'.

        :param drug_entry_ids: The drug entry ID options.
        :return: The constructed options.
        """
        return "+".join(drug_entry_ids)



import setuptools as st
import re


requirements = [
    'docopt',
    'requests'
]


def _readme() -> str:
    with open('README.rst') as readme_file:
        return readme_file.read()


def _get_version() -> str:
    with open('src/kegg_pull/__init__.py', 'r') as fd:
        version: str = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                            fd.read(), re.MULTILINE).group(1)
    if not version:
        raise RuntimeError('Cannot find version information')

    return version


def _readme():
    with open('README.rst') as readme_file:
        return readme_file.read()


st.setup(
    name='kegg_pull',
    version=_get_version(),
    package_dir={'': 'src'},
    packages=st.find_packages('src', exclude=['dev', 'docs']),
    install_requires=requirements,
    entry_points={'console_scripts': ['kegg_pull = kegg_pull.__main__:main']},
    author='Erik Huckvale',
    author_email='edhu227@g.uky.edu',
    url='https://github.com/MoseleyBioinformaticsLab/KEGGpull',
    long_description_content_type='text/x-rst',
    long_description=_readme()
)



import logging as l
import typing as t


def split_comma_separated_list(list_string: str) -> list:
    items: list = list_string.split(',')

    if '' in items:
        l.warning(f'Blank items detected in the comma separated list: "{list_string}". Removing blanks...')
        items = [item for item in items if item != '']

    return items


def get_molecular_attribute_args(args: dict) -> tuple:
    formula: str = args['--formula']
    exact_mass: list = args['--exact-mass']
    molecular_weight: list = args['--molecular-weight']

    # exact_mass and molecular_weight will be [] (empty list) if not specified in the commandline args
    if exact_mass:
        exact_mass: t.Union[float, tuple] = _get_range_values(range_values=exact_mass, value_type=float)
    else:
        exact_mass = None

    if molecular_weight:
        molecular_weight: t.Union[int, tuple] = _get_range_values(range_values=molecular_weight, value_type=int)
    else:
        molecular_weight = None

    return formula, exact_mass, molecular_weight


def _get_range_values(range_values: t.Union[int, float, tuple], value_type: type) -> t.Union[int, float, tuple]:
    if len(range_values) == 1:
        [val] = range_values

        return value_type(val)
    elif len(range_values) == 2:
        [min_val, max_val] = range_values

        return value_type(min_val), value_type(max_val)
    else:
        raise ValueError(
            f'Range can only be specified by two values but {len(range_values)} values were provided: '
            f'{", ".join(str(range_value) for range_value in range_values)}'
        )


class staticproperty(staticmethod):
    def __get__(self, *_):
        return self.__func__()
